#include <v_lib.h>
#include "rgcache.h"

typedef enum
{
	fg_constant = 1,
	fg_uniform = 2,
	fg_gaussian = 3,
	fg_iterators = 99,
	fg_linear_iterator = 100,
	fg_log10_iterator = 101
} float_generator_type;

/* 
	TODO: extend 'for' with 'logbase', can easily perform arbitrary base interpolation
 */

typedef struct float_generator_s
{
	float_generator_type	type;
	char	*name;
	vboolean finished; /* iterations complete */
	union
	{
		vec_t constant;
		struct
		{
			rgcache_p rgc;
			vec_t min, max;
		} uniform;
		struct
		{
			rgcache_p rgc; /* must be at the same place */
			vec_t mean, stdev;
		} gaussian;
		struct
		{
			u32_t steps, idx; /* required for bidirectional iterators */
			vec_t from, to, val, delta;
		} ilinear;
		struct
		{
			u32_t steps, idx; /* required for bidirectional iterators */
			vec_t log10from, log10to, log10val, log10delta;
		} ilog10;
	} u;
	vboolean (*next)(struct float_generator_s *self,vec_t *ret_val);
	void (*reset)(struct float_generator_s *self);
	void (*destroy)(struct float_generator_s *self);
} float_generator_t, *float_generator_p;

static vboolean
fg_next_constant(struct float_generator_s *self,vec_t *ret_val)
{
	*ret_val = self->u.constant;
	return true;
}

static vboolean
fg_next_uniform(struct float_generator_s *self,vec_t *ret_val)
{
	*ret_val = rgcNext(self->u.uniform.rgc);
	return true;
}

static vboolean
fg_next_gaussian(struct float_generator_s *self,vec_t *ret_val)
{
	*ret_val = rgcNextGaussian(self->u.uniform.rgc);
	return true;
}

static vboolean
fg_next_linear(struct float_generator_s *self,vec_t *ret_val)
{
	if(self->finished)
		return false;
	*ret_val = self->u.ilinear.val;
	self->u.ilinear.val += self->u.ilinear.delta;
	self->u.ilinear.idx++;
	if(self->u.ilinear.idx > self->u.ilinear.steps)
		self->finished = true;
	return true;
}

static vboolean
fg_next_log10(struct float_generator_s *self,vec_t *ret_val)
{
	if(self->finished)
		return false;
	*ret_val = (vec_t)pow(10, self->u.ilog10.log10val);
	self->u.ilog10.log10val += self->u.ilog10.log10delta;
	self->u.ilog10.idx++;
	if(self->u.ilog10.idx > self->u.ilog10.steps)
		self->finished = true;
	return true;
}

static void
fg_dummy_reset(struct float_generator_s *self)
{
	return;
}

static void
fg_iterator_reset(struct float_generator_s *self)
{
	self->u.ilinear.idx = 0;
	self->u.ilinear.val = self->u.ilinear.from;
	self->finished = false;
	return;
}

static void
fg_generic_destroy(struct float_generator_s *self)
{
	if(!self)
		return;
	if(self->name)
		memFree(self->name);
	memFree(self);
}

static void
fg_random_destroy(struct float_generator_s *self)
{
	if(self && self->u.uniform.rgc)
		rgcDestroy(self->u.uniform.rgc);
	fg_generic_destroy(self);
}

static vboolean
parseConstantGen(vhandle_t lex,vp_stack_p storage)
{
	CALLER_NAME(parseConstantGen)
	float_generator_p fg = (float_generator_p)memAllocZ(sizeof(float_generator_t));
	fg->type = fg_constant;
	fg->name = strDup(lexNext(lex));
	fg->finished = false;
	fg->next = fg_next_constant;
	fg->reset = fg_dummy_reset;
	fg->destroy = fg_generic_destroy;
	if(lexNextFloat(lex,&(fg->u.constant)))
	{
		vpsPush(storage, fg);
		return true;
	}
	PARSE_ERROR(lex, "generic error");
	fg_generic_destroy(fg);
	return false;
}

static vboolean
parseUniformGen(vhandle_t lex,vp_stack_p storage)
{
	CALLER_NAME(parseUniformGen)
	float_generator_p fg = (float_generator_p)memAllocZ(sizeof(float_generator_t));
	fg->type = fg_uniform;
	fg->name = strDup(lexNext(lex));
	fg->finished = false;
	fg->next = fg_next_uniform;
	fg->reset = fg_dummy_reset;
	fg->destroy = fg_random_destroy;
	if(lexNextFloat(lex,&(fg->u.uniform.min)) &&
		lexNextFloat(lex,&(fg->u.uniform.max)))
	{
		fg->u.uniform.rgc = rgcReset(NULL, fg->u.uniform.min, fg->u.uniform.max, 0, 1);
		vpsPush(storage, fg);
		return true;
	}
	PARSE_ERROR(lex, "generic error");
	fg_generic_destroy(fg);
	return false;
}

static vboolean
parseGaussianGen(vhandle_t lex,vp_stack_p storage)
{
	CALLER_NAME(parseGaussianGen)
	float_generator_p fg = (float_generator_p)memAllocZ(sizeof(float_generator_t));
	fg->type = fg_uniform;
	fg->name = strDup(lexNext(lex));
	fg->finished = false;
	fg->next = fg_next_gaussian;
	fg->reset = fg_dummy_reset;
	fg->destroy = fg_random_destroy;
	if(lexNextFloat(lex,&(fg->u.gaussian.mean)) &&
		lexNextFloat(lex,&(fg->u.gaussian.stdev)))
	{
		fg->u.uniform.rgc = rgcReset(NULL, 0, 1, fg->u.gaussian.mean, fg->u.gaussian.stdev);
		vpsPush(storage, fg);
		return true;
	}
	PARSE_ERROR(lex, "generic error");
	fg_generic_destroy(fg);
	return false;
}

static vboolean
parseIterator(vhandle_t lex,vp_stack_p storage)
{
#define	MAXSTEPS	65536
	CALLER_NAME(parseIterator)
	vec_t from, to;
	s32_t steps;
	float_generator_p fg = 
		(float_generator_p)memAllocZ(sizeof(float_generator_t));
	fg->name = strDup(lexNext(lex));
	fg->finished = false;
	fg->reset = fg_iterator_reset;
	fg->destroy = fg_generic_destroy;
	if(!lexNextFloat(lex,&from))
		goto parse_error;
	if(!lexNextIs(lex,"to"))
		goto parse_error;	
	if(!lexNextFloat(lex,&to))
		goto parse_error;   
	if(!lexNext(lex))
		goto parse_error;
	if(lexLastWas(lex, "steps"))
	{
		if(!lexNextInteger(lex, &steps) || (u32_t)steps > MAXSTEPS)
			goto parse_error;
		fg->type = fg_linear_iterator;
		fg->next = fg_next_linear;
		fg->u.ilinear.from = from;
		fg->u.ilinear.to = to;
		fg->u.ilinear.steps = (u32_t)steps;
		fg->u.ilinear.idx = 0;
		fg->u.ilinear.val = fg->u.ilinear.from;
		fg->u.ilinear.delta = (fg->u.ilinear.to - fg->u.ilinear.from) 
			/ fg->u.ilinear.steps;
	}
	else if(lexLastWas(lex, "logsteps"))
	{
		if(!lexNextInteger(lex, &steps) || (u32_t)steps > MAXSTEPS)
			goto parse_error;
		fg->type = fg_log10_iterator;
		fg->next = fg_next_log10;
		fg->u.ilog10.log10from = (vec_t)log10(from);
		fg->u.ilog10.log10to = (vec_t)log10(to);
		fg->u.ilog10.steps = (u32_t)steps;
		fg->u.ilog10.idx = 0;
		fg->u.ilog10.log10val = fg->u.ilog10.log10from;
		fg->u.ilog10.log10delta = 
			(fg->u.ilog10.log10to - fg->u.ilog10.log10from) / steps;
	}
	else
		goto parse_error;
	vpsPush(storage, fg);
	return true;
parse_error:;
	PARSE_ERROR(lex, "generic error");
	fg_generic_destroy(fg);
	return false;
}

/*
 * Returns number of generators stored in 'store'.
 */
unsigned V_GLOBAL
parseGenerators(vhandle_t lex,vp_stack_p store)
{
	u32_t	i, stored = 0;
	float_generator_p fg, newg;
	CALLER_NAME(parseGenerators)
	if(!lexLastWas(lex,"generators"))
		conlPrintf(LOG_WARNING,"%s(): prev keyword was strange...\n",caller_name);
	if(!lexNextIs(lex,"{"))
	{
		SYN_EXPECTED(lex,"{");
		return stored;
	}
	/* TODO: lexNextMatches() */
	while(!lexNextIs(lex, "}"))
	{
		if(!lexLast(lex))
		{
			SYN_UNEXPECTED_EOF(lex);
			break;
		}
		if(lexLastWas(lex, "constant"))
		{
			if(!parseConstantGen(lex, store))
				break;
		}
		else if(lexLastWas(lex, "uniform"))
		{
			if(!parseUniformGen(lex, store))
				break;
		}
		else if(lexLastWas(lex, "gaussian"))
		{
			if(!parseGaussianGen(lex, store))
				break;
		}
		else if(lexLastWas(lex, "iterator"))
		{
			if(!parseIterator(lex, store))
				break;
		}
		else
		{
			SYN_EXPECTED(lex, "constant|uniform|gaussian|iterator");
			break;
		}
		if(!lexNextIs(lex, ";"))
		{
			SYN_EXPECTED(lex, ";");
			break;
		}
		/*
		 * Check for duplicate names
		 */
		newg = (float_generator_p)vpsGet(store, vpsCount(store) - 1);
		for(i = 0;i < vpsCount(store) - 1;i++)
		{
			fg = vpsGet(store, i);
			if(!strcmp(newg->name, fg->name))
			{
				conlPrintf(LOG_WARNING,
					"%s(): in \"%s\", line %u: duplicate name \"%s\", discarded.\n",
					caller_name, lexSourceName(lex), lexLastLine(lex),
					newg->name);
				vpsPop(store, &newg);
				newg->destroy(newg);
			}
		}
	}
	return stored;
}

vhandle_t in;

vp_stack_t floatgens = TYPED_STACK_EMPTY_INITIALIZER;

static void
clean_floatgens()
{
	float_generator_p g;
	while(vpsPop(&floatgens,(void_p*)&g))
		g->destroy(g);
	vpsClear(&floatgens);
} 

int main(int argc, char *argv[])
{
	CALLER_NAME(main);
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	in = lexParseFile("test.fcic","{}()[];");
	lexSetInteger(in,LEX_COMMENT_CHAR,'#');
	lexEnable(in,LEX_ONE_CHAR_COMMENTS);
	while(lexNext(in))
	{
		if(lexLastWas(in, "generators"))
		{
			parseGenerators(in, &floatgens);
		}
	}
    printf("parsed %u generators\n", vpsCount(&floatgens));
    if(vpsCount(&floatgens))
    {
    	int i;
    	vec_t f;
    	float_generator_p g;
    	g = (float_generator_p)vpsGet(&floatgens,0);
    	for(i=0;i<10;i++)
    	{
    		if(!g->next(g,&f))
    			break;
    		printf("%g\n",f);
    	}
    }
    clean_floatgens();
	lexDestroy(in);
	memPrintStats();
	return 0;
}
